#!/data/oir/bin/ksh
#
# Generate starbase tables of:
#
#   1) All the keywords in all FITS files     --> ${dirname}_keyval.dat
#   2) All Object files                       --> ${dirname}_objlist.dat
#   3) All Skyfiles & Certification values    --> ${dirname}_skylist.dat
#   3) Objects and distances to each Sky file --> ${dirname}_skydist.dat
#
# Number (3) uses keywords FIBER_X and FIBER_Y to compute the distance
# between an object and each of the sky files in turn.
#
# Also, assumes the UNIX IRAF environment is set up (for imhbase, and all
# the other IRAF utilities).
#
#  
#
# Produces a <dir>_skydist.db starbase file that looks like (example):
# HDRFILE   Skyfile   FIBER_X FIBER_Y SkyX    SkyY    SkyDist SkyR
# --------- --------- ------- ------- ------- ------- ------- ----
# obj.fits  sky4.fits  10.000  10.000  70.000  11.000    60.0  1.8
# obj.fits  sky3.fits  10.000  10.000 -50.000  40.000    67.1  2.2
# obj2.fits sky4.fits  20.000  60.000  70.000  11.000    70.0  1.8
# obj2.fits sky3.fits  20.000  60.000 -50.000  40.000    72.8  2.2
#
#

# ========================================================================
#       Setup the Korn Shell envirnment for variables and parameter files
# ========================================================================
export home=${irafhome:-$HOME/unixiraf}
export DEFAULT_DIR=${iraf}../iraf.pkg/hectospec/
export PATH=/data/oir/unixiraf12.1/bin:$PATH
#export PATH=/data/oir/starbase2.7.8:$PATH

. /data/oir/unixiraf12.1/scripts/ui_setup.ksh 

. $ENVIRONMENT

# ========================================================================
#       Get runtime parameters from script parameter file
# ========================================================================
#

verbose=0
files="\*.fits"
destdir="skysub/"
mpjobs=12
nskys=0
fiberdist=0.
logfile=skyproc.results
skycert=4.
skystdlim=0.2
skysame=yes         ## NEW 2004-04-15 WFW
emline=1
uparmdir=""
#echo files="$files"

###  "i:d:m:n:f:e:w:l:r:s:u:v:h"

while getopts "i:d:m:n:e:l:r:s:u:v:c:h" fval ; do
#   echo $fval
   case $fval in
     c) [[ $OPTARG == yes || $OPTARG == 1 ]] && skysame=1
        [[ $OPTARG == no  || $OPTARG == 0 ]] && skysame=0
        ;;
     i) files=$OPTARG
	if (( verbose > 1 )) then
	    echo files=$files
	    fi;;
     d) destdir=$OPTARG
	if (( verbose > 1 )) then
	    echo destdir=$destdir
	    fi;;
#     e) emline=$OPTARG
#	if (( verbose > 1 )) then
#	    echo emline=$emline
#	    fi;;
#     w) emparm=$OPTARG
#	if (( verbose > 1 )) then
#	    echo emparm=$emparm
#	    fi;;
     m) mpjobs=$OPTARG
	if (( verbose > 1 )) then
	    echo mpjobs=$mpjobs
	    fi;;
     n) nskys=$OPTARG
	if (( verbose > 1 )) then
	    echo nskys=$nskys
	    fi;;
     f) fiberdist=$OPTARG
	if (( verbose > 1 )) then
	    echo fiberdist=$fiberdist
	    fi;;
     l) logfile=$OPTARG
	if (( verbose > 1 )) then
	    echo logfile=$logfile
	    fi;;
     r) skycert=$OPTARG
	if (( verbose > 1 )) then
	    echo skycert=$skycert
	    fi;;
     s) skystdlim=$OPTARG
	if (( verbose > 1 )) then
	    echo skystdlim=$skystdlim
	    fi;;
     u) uparmdir=$OPTARG
	if (( verbose > 1 )) then
	    echo uparmdir=$uparmdir
	    fi;;
     v) verbose=$OPTARG
	    echo verbose=$verbose;;
     h) print  "Usage: $(basename $0) [options]"
        printf  "  -i %s\tinput file list\n" $files
        printf  "  -d %s\tdestination directory for output files\n" $destdir
        printf  "  -c %d\tUse same chip for skys? (1=yes)\n" $skysame
        printf  "  -n %d\tMax number of skys to coadd\n" $nskys
        printf  "  -f %.2f\tMax fiber distance to sky\n" $fiberdist
        printf  "  -l %s\tName of output summary file\n" $logfile
        printf  "  -r %.2f\tMax R value for certified sky\n" $skycert
        printf  "  -s %.2f\tMax std of difference from avg sky\n" $skystdlim
        printf  "  -m %d\tMax number of multiprocessor jobs in parallel\n" $mpjobs
        printf  "  -u %s\tName of uparm directory\n" $uparmdir
        printf  "  -v 1\tverbose - output summary on stderr\n" \
                   $(if (( verbose > 0 )) then echo yes; else echo no;fi)
        printf  " -v 2\tvery verbose: show each ap line change\n"\
                   $(if (( verbose > 1 )) then echo yes; else echo no;fi)
        printf  "  -h\t\tprint this usage and exit\n"
        printf  "\n"
        exit 1;;
#    \?) print -u2 "Unknown option $OPTARG"; exit 1;;
     :) print -u2 "-$OPTARG requires an argument"; exit 1;;
   esac
done



if (( mpjobs > 4096 )) then
    mpjobs = 4096
elif (( mpjobs < 1 )) then
    mpjobs = 1
fi

# Make sure output directory is there
destdir=`pwd`/$destdir
if [[ -d $destdir ]] then
    x=0
else
    mkdir -p $destdir
fi

trap exit ERR
export archive=$irafpkg/saotdc/pkg/archive/
export imextn="oif:imh fxf:fits,fit,fts plf:pl qpf:qp stf:hhh,??h"

skymethod="SKYSUB"

REV=2.1s

tmp="/tmp/$(basename $0).$$"/
mkdir -p $tmp

tmpfileA=${tmp}$$A
tmpfileB=${tmp}$$B
tmpskylist=${tmp}$$.skyhdr

procpid=$$

exitval=0

trap 'print -u2 -f "ERROR $? in $0 at line $LINENO\n"; exitval=1; exit' ERR
trap "/bin/rm -rf $tmp" EXIT

# Put parameter files in destination directory for use by 
# unixiraf tasks

uparmdir=$destdir

# Initialize parameter files from hectospec package if not in uparm directory
# and copy to destination directory for use by subroutines
if test !(-a ${UPARM}hecskxcso.par)
    then
    cp -p ${DEFAULT_DIR}skxcsao.par ${UPARM}hecskxcso.par
    fi
cp -p $uparmdir/hecskxcso.par $destdir/xcsao.par

if test !(-a ${UPARM}hecskemso.par)
    then
    cp -p ${DEFAULT_DIR}skemsao.par ${UPARM}hecskemso.par
    fi
cp -p $uparmdir/hecskemso.par $destdir/emsao.par

if test !(-a ${UPARM}hecskexpr.par)
    then
    cp -p ${DEFAULT_DIR}skexpr.par ${UPARM}hecskexpr.par
    fi
cp -p $uparmdir/hecskexpr.par $destdir/imexpr.par

if test !(-a ${UPARM}hecskstak.par)
    then
    cp -p ${DEFAULT_DIR}skstack.par ${UPARM}hecskstak.par
    fi
cp -p $uparmdir/hecskstak.par $destdir/imstack.par

if test !(-a ${UPARM}hecskarih.par)
    then
    cp -p ${DEFAULT_DIR}skarith.par ${UPARM}hecskarih.par
    fi
cp -p $uparmdir/hecskarih.par $destdir/imarith.par

if test !(-a ${UPARM}hecskcome.par)
    then
    cp -p ${DEFAULT_DIR}skcombine.par ${UPARM}hecskcome.par
    fi
cp -p $uparmdir/hecskcome.par $destdir/imcombine.par

if test !(-a ${UPARM}hecskcome.par)
    then
    cp -p ${DEFAULT_DIR}skstatistics.par ${UPARM}hecskstas.par
    fi
cp -p $uparmdir/hecskstas.par $destdir/imstatistics.par

export UPARM=$uparmdir
# Look for parameter files in destination directory
if (( verbose > 0 )) then
    echo Parameter files are in $UPARM
    fi

# String prefacing output .db files - default is the directory name.
prefix=${prefix:-$(basename $(pwd))_}
[[ $prefix == "NONE" ]] && prefix=""

#echo skyproc: prefix is $prefix

if [[ -z "$logfile" ]] then
    logfile="/dev/null"
    fi

if [[ $logfile != "-" && $logfile != "STDOUT" ]] then
    exec 1>> $logfile
    fi

if (( verbose > 0 )) then
    printf "\n%s rev %s start at $(date)\n\n" $(basename $0) $REV
    printf "  Nskys = %d\n  FiberDist = %.1f\n  SkyCert <= %.1f\n" \
              $nskys $fiberdist $skycert
    printf "  SkyStdLim = %.2f\n" $skystdlim
    printf "  DB Files prefix = %s\n" $prefix
    printf "  DestDir = %s\n\n" $destdir
    fi


# Make list of 1-D spectrum files to be processed from input argument
#files=`echo $files | tr "," " "`
#files=`ls $files`
#if (( verbose > 1 ))
#    then
#    echo processing files $files
#    fi

# Generate master table of keywords; works for FITS or IRAF files
gethead -tha IMAGETYP OBJECT FIBER FIBER_X FIBER_Y $files |
  trim   FILENAME=HDRFILE -a                                |
  compute 'if(OBJECT ~ /_-1$/) {OBJECT = "UNUSED"; IMAGETYP = "UNUSED" }
           if(OBJECT ~ /_0$/ ) {OBJECT = "SKY"}'            |
  sorttable HDRFILE >${prefix}keyval.db

objlist=${prefix}objlist.db
if (( verbose > 0 ))
    then
    echo Master object list is in $objlist
    fi

# Pull all [object and not sky] files into a master list
<${prefix}keyval.db \
      row 'IMAGETYP == "OBJECT" && toupper(OBJECT) !~ "SKY"   \
           && toupper(OBJECT) !~ /(UNUSED|PARKED)/'             |
      column HDRFILE OBJECT FIBER FIBER_X FIBER_Y             \
>$objlist 

# create a sky file db
row 'IMAGETYP == "OBJECT" && toupper(OBJECT) ~ "SKY"' <${prefix}keyval.db |
column HDRFILE  >${tmp}sky.db

# create the sky file list (not a .db file).
headoff HDRFILE <${tmp}sky.db >$tmpskylist
cp $tmpskylist sky.list

nsky=$(headoff HDRFILE <${tmp}sky.db | wc -l)

# compute master sky average, then ratio each against the average to pick
# up abberant skys to remove from the master sky list

tmprat=${tmp}skyratio
sed <$tmpskylist -e 's/\.sky_.*\./.ratio./' -e "s,^,$tmp," >$tmprat

imcombine @$tmpskylist ${tmp}skyav rejmask="" combine="average"          \
       reject="avsigclip"   outtype="real"  scale="none"   zero="none"   \
       weight="!EXPTIME"  mclip="yes"  lsigma=3. hsigma=3. sigscale=0.1  \
       logfile=""
imarith @$tmpskylist / ${tmp}skyav @$tmprat
imstatistics @$tmprat fields=stddev format=no   |
fldtotable skyavstd > ${tmp}skyav.db
pastetable ${tmp}sky.db ${tmp}skyav.db     |
sorttable HDRFILE >${tmp}skyavstd.db

/bin/rm $(cat $tmprat)

seqno=0

# Select emission line to use for sky/object scaling
if (( emparm == 1 )) then
    linecol=lht0$emline
elif (( emparm == 2 )) then
    linecol=lwd0$emline
else
    linecol=leqw0$emline
fi
echo Scaling spectra by $linecol

nsplit=$(( nsky / mpjobs ))
if (( nsplit * mpjobs < nsky )) 
    then
    nsplit=$(( nsplit + 1 ))
    fi
if (( verbose > 0 )) then
    echo $nsky sky files: forking $mpjobs subshells to each run emsao on $nsplit
    fi
echo $tmpskylist
split -l $nsplit $tmpskylist ${tmp}skypart.

for skpart in ${tmp}skypart.*; do
  seqno=$((seqno + 1))
  (
    xit=0
    tdir=${tmp}skdir.$seqno/
    mkdir -p $tdir 2>/dev/null
    cp $uparmdir/xcsao.par $tdir 
    export UPARM=$tdir

    xcsao @${skpart}  report_mode=-8 displot=no mode=q |
    column specid=HDRFILE xcr=SkyR >${skpart}.db

    xit=$?
    [[ $xit != 0 ]] && {
      print "$(basename $0): Error=$xit in xcsao, seq=$seqno \?\n"
    }
    exit $xit
 ) &  # END FORKED SUBSHELL
done  
# wait for the subprocesses to finish
wait

# Pull the sky tables all together
if [[ $(echo ${tmp}skypart*.db | wc -w) -gt 1 ]]; then
  uniontable ${tmp}skypart*.db | sorttable HDRFILE >${tmp}skylist.db
else
  sorttable HDRFILE < ${tmp}skypart.aa.db  >${tmp}skylist.db
fi

/bin/rm -rf ${tmp}skypart* ${tmp}skdir.*

# add in the avstd field from above
jointable -j HDRFILE ${tmp}skylist.db ${tmp}skyavstd.db >${prefix}skylist.db

# Pull join (cprodtable) all certified skyfiles to object list, compute 
# distance from each object to each sky, leave sorted by object and
# then sky distances.

cprodtable  $objlist                                      \
   <(jointable -j HDRFILE ${prefix}keyval.db ${prefix}skylist.db     \
      | column HDRFILE FIBER FIBER_X FIBER_Y SkyR skyavstd=SkyAvStd  \
      | row '(SkyR ~ /^ *$/ || SkyR <= '$skycert') && SkyAvStd < '$skystdlim \
    )                                                                \
    | column HDRFILE_1=HDRFILE HDRFILE_2=Skyfile                     \
          FIBER_1=FIBER FIBER_X_1=FIBER_X FIBER_Y_1=FIBER_Y          \
          FIBER_2=SkyFIBER FIBER_X_2=SkyX FIBER_Y_2=SkyY             \
             SkyDist SkyR SkyAvStd                                   \
    | compute 'SkyDist=sprintf("%.1f",                                    \
                         sqrt(square(FIBER_X-SkyX) + square(FIBER_Y-SkyY)))'\
    | sorttable HDRFILE SkyDist:n                                         \
>${prefix}skydist.db

# just to see the final table
if (( verbose > 1 )) then
    justify <${prefix}skydist.db                                          \
	HDRFILE,7 Skyfile,7 FIBER,4 FIBER_X,8 FIBER_Y,8                   \
	SkyFIBER,6 SkyX,8 SkyY,8 SkyDist,7 SkyR,4 
    fi

if (( verbose > 0 )) then
    printf "\nElapsed time after object & sky tables creation: $SECONDS seconds\n"
    times
    fi

# Run emsao on list of object files to get emission line characteristics and
# add value of appropriate column to table.
pset emsao report_mode=-4 displot=no esmooth=2
pset emsao emsearch=searchO5577.dat emlines=hectosky.dat linedir='rvsao$lib'

headoff HDRFILE <$objlist >$tmpfileA
seqno=0
nobj=$(headoff HDRFILE <$objlist | wc -l)
nsplit=$(( nobj / mpjobs ))
if (( nsplit * mpjobs < nobj )) 
    then
    nsplit=$(( nsplit + 1 ))
    fi
if (( verbose > 0 )) then
    echo $nobj objects: forking $mpjobs subshells to each run emsao on $nsplit
    fi
split -l $nsplit $tmpfileA ${tmp}objpart.

for objpart in ${tmp}objpart.*; do
  seqno=$((seqno + 1))
  (
    xit=0
    tdir=${tmp}objdir.$seqno/
    mkdir -p $tdir 2>/dev/null
    cp $destdir/emsao.par $tdir
    export UPARM=$tdir

    emsao @$objpart report_mode=-4 displot=no esmooth=2 mode=q  | \
    column specid=HDRFILE $linecol=objnorm  >$tmpfileB.$seqno >${objpart}.db


    xit=$?
#    echo $objpart completed: error code=$xit.
    if (( xit != 0 )) then
	print "$(basename $0): Error=$xit in emsao, seq=$seqno \?\n"
	fi
    exit $xit
  ) &  # END FORKED SUBSHELL
done  

# wait for the subprocesses to finish
wait

ntab=$(ls ${tmp}objpart*.db | wc -l)

# Pull the sky tables all together
if (( ntab > 1 )) then
  uniontable ${tmp}objpart*.db | sorttable HDRFILE >$tmpfileB
else
  sorttable HDRFILE <${tmp}objpart.aa.db >$tmpfileB
fi

#echo $ntab tables of object emsao norm results combined into $tmpfileB

nobja=$(headoff HDRFILE <$objlist | wc -l)
#echo $nobja objects in $objlist
if (( nobja < 1 )) then
   printf "No rows in $objlist to be combined?\n"
   exit 1
fi
nobjb=$(headoff HDRFILE <$tmpfileB | wc -l)
#echo $nobjb objects in $tmpfileB
if (( nobjb < 1 )) then
   printf "No rows in $tmpfileB to be combined?\n"
   exit 1
fi
if (( nobjb != nobja )) then
   printf "$nobja rows in $tmpfileA, $nobjb in $tmpfileB\n"
   exit 1
#else
#   echo $nobja rows in $tmpfileA, $nobjb in $tmpfileB
fi

jointable -j HDRFILE $objlist $tmpfileB >$tmpfileA
nobj=$(headoff HDRFILE <$tmpfileA | wc -l)
#echo $nobj object files tabulated in $tmpfileA

if (( nobj < 1 )) then
   mv $tmpfileB emsao.db
   printf "No rows in jointable?\n"
   exit 1
fi
mv $tmpfileA $objlist
/bin/rm -rf ${tmp}objpart.* ${tmp}objdir.* $tmpfileB.*

if (( verbose > 0 )) then
    printf "\nElapsed time after object emission line fit: $SECONDS seconds\n"
    times
    fi

#
# Create entries in table for sky summation list for each object
#
seqno=0
waitjob=0
/bin/rm -f ${tmp}nskys.*

for obj in $(headoff <$objlist HDRFILE)
 do
   seqno=$((seqno + 1))
 (
   tmpA=${tmpfileA}.${seqno}
# Select all skyfiles that are certified and within FIBERDIST of object
   trim <${prefix}skydist.db    |
   table -h 'HDRFILE == "'$obj'" && (SkyR <= '$skycert' || SkyR ~ /^ *$/) \
                 && ('$fiberdist' == 0 ||                                 \
                     SkyDist <= '$fiberdist') {print Skyfile}'            \
    >$tmpA

# Limit to the max allowed number of skys. This chooses the closet ones 
# because the input table was sorted secondarily by SkyDist.
  if (( nskys > 0 )) then
     head -${nskys} $tmpA      |
     sort >${tmpskylist}.$seqno
  else
     sort <$tmpA >${tmpskylist}.$seqno
  fi

# ERROR if no sky files left after selection
  if [[ ! -s ${tmpskylist}.$seqno ]]; then    
    print -u2 -f "\nWARNING:" 
    print -u2 -f " File %s has no sky files within the %.3f mm limit!\n"   \
       $obj $fiberdist
    exitval=1
    continue
  fi
  nskyfiles=$(wc -l <${tmpskylist}.$seqno)
  skylist=$(tr '\012' '+' <${tmpskylist}.$seqno)
  skylist=${skylist%+}
  aplist=$(gethead -a @${tmpskylist}.$seqno aperture | 
            cut -f 2 -d\  | 
            tr '\012' '+')
  printf "%s\t%s\t%d\t%s\n" $obj ${skylist%_} $nskyfiles ${aplist%+} \
       >>${tmp}nskys.$seqno
 ) &
  if (( seqno - waitjob >= mpjobs )) then
      waitjob=$((waitjob + 1))
      wait %${waitjob}
  fi
 done 

 wait 

 cat ${tmp}nskys.*                                |
 fldtotable HDRFILE Skyfilelist Nskys Apertures   | 
 sorttable HDRFILE >$tmpfileB

jointable -j HDRFILE $objlist $tmpfileB  >$tmpfileA

mv $tmpfileA $objlist

/bin/rm -f ${tmp}nskys.*   ${tmpskylist}.*

if (( verbose > 1 )) then
    printf "\nSet of skys for object files:\n"
    justify <$objlist HDRFILE Nskys Apertures
    fi

#
# Create sorted, unique set of skyfile lists with sky normalization value
#
pset emsao report_mode=-4 displot=no esmooth=2
pset imcombine project=no logfile=""

skynum=0

headoff Skyfilelist <$objlist | sort -u >${tmp}skyfilelist
nskycomb=$(headoff Skyfilelist <$objlist | sort -u |  wc -l)

seqno=0
nsplit=$(( nskycomb / mpjobs ))
if (( nsplit * mpjobs < nskycomb ))
    then
    nsplit=$(( nsplit + 1 ))
    fi
if (( verbose > 0 )) then
    echo $nskycomb sky combinations: Forking $mpjobs subshells to each process $nsplit
    fi
split -l $nsplit ${tmp}skyfilelist ${tmp}skyflist.

for skylist in ${tmp}skyflist.* ; do
  seqno=$((seqno + 1))
  skynum=0
  if [[ -e ${skylist}.db ]] then
    /bin/rm -rf ${skylist}.db
    fi

# fork off for multiproc
  (
    xit=0
    tdir=${skylist}.$seqno/
    mkdir -p $tdir 2>/dev/null
    cp $destdir/emsao.par  $destdir/imcombine.par $tdir
    export UPARM=$tdir
#
# Get sky fiber list from sky file names
#
    cat $skylist   |
    while read skyfiblist; do
      skynum=$((skynum + 1))
      nskyfibers=$(echo $skyfiblist | tr "+" "\012" | wc -l)
      skyfibers="$(    \
       (for file in $(echo $skyfiblist | tr '+' ' '); do
          table -h <${prefix}keyval.db 'HDRFILE == "'$file'" {print FIBER}'
        done) | tr "\012" "_")"

#      extn=${skyfiblist##*\.}
#      extn=${extn%%+}
      extn=".fits"
      skysumfile=${tmp}${skyfibers}.${procpid}${extn}
      shortname=${tmp}skyset_${seqno}_${skynum}${extn}
#  shortname=skyset_${seqno}_${skynum}{$extn}
#      /bin/rm -f $skysumfile $shortname
      ls `echo $skyfiblist | tr "+" " "` > ${tdir}skyfiles
      imcombine @${tdir}skyfiles $shortname mode=q
      skynorm=$(emsao $shortname mode=q |  headoff $linecol)
      xit=$?
      if (( xit != 0 )) then
        print "$(basename $0): Error=$xit in emsao, seq=$seqno \?\n"
	fi
      printf "%s\t%s\t%s\t%s\t%s\t%s\n" $skyfiblist $skysumfile $shortname \
         "$nskyfibers" $skyfibers $skynorm >>${skylist}.db
    done
  ) &
done

wait

cat ${tmp}skyflist*.db |
fldtotable Skyfilelist Skysumfile Skyabbrev Nskyfibers Skyfibers skynorm |
sorttable Skyfilelist >$tmpfileA

/bin/rm -rf ${tmp}skyfilelist ${tmp}skyflist*

jointable -j Skyfilelist <(sorttable Skyfilelist <$objlist) \
                          $tmpfileA                                     |
 column -a ratio                                                        |
 compute 'ratio=sprintf("%.6f", objnorm/skynorm)'  >$tmpfileB

sorttable HDRFILE <$tmpfileB >$objlist

if (( verbose > 0 )) then
    printf "\nElapsed time after summed sky emission line fit: $SECONDS seconds\n"
    times
    fi

# Subtract off scaled sky
#typeset -Z5 seqno=0
seqno=0

if (( verbose > 1 )) then
    printf "\n"
    fi

columns="HDRFILE Skysumfile Skyabbrev Nskyfibers Skyfibers Apertures objnorm skynorm ratio"

nxtjob=0
waitjob=0

headoff $columns <$objlist |&
while read -p $columns; do
  seqno=$(($seqno + 1))

# Fork off up to $mpjobs subshells into the background, waiting for
# one before doing the next
 (  
  trap \
   'print -f -u2 "ERROR $? in $0 seq $seqno at line $LINENO\n"; exit 100' ERR

  if (( verbose > 1 )) then
    print -u2 -f "Forking skysub job %d for %s\n" $seqno $HDRFILE
    fi

  exec 1> ${tmp}logfile.$seqno 
#  exec 1> /tmp/skyproclog.$seqno 

  extn=$(expr $HDRFILE : '.*\(\..*\)$' \| $HDRFILE)
  outbase=$(basename $HDRFILE $extn)
  outfile=${tmp}${outbase}.sub${extn}
  destfile=${destdir}${outbase}.ms${extn}
  tmpdestfile=${tmp}${outbase}.ms${extn}
  if (( verbose > 1 )) then
    if (( linecol == 1 )) then
	printf "%s: Obj O(5577)= %.1f, Sky O(5577)= %.1f, ratio= %.3f\n"\
           $HDRFILE $objnorm $skynorm $ratio
    elif (( linecol == 2 )) then
	printf "%s: Obj O(6300)= %.1f, Sky O(6300)= %.1f, ratio= %.3f\n"\
           $HDRFILE $objnorm $skynorm $ratio
    elif (( linecol == 3 )) then
	printf "%s: Obj OH(7341)= %.1f, Sky OH(7341)= %.1f, ratio= %.3f\n"\
           $HDRFILE $objnorm $skynorm $ratio
    else
	printf "%s: Obj OH(8399)= %.1f, Sky OH(8399)= %.1f, ratio= %.3f\n"\
           $HDRFILE $objnorm $skynorm $ratio
    fi
    printf "imexpr %s = %s - (%s * %s)\n" $outfile $HDRFILE $Skyabbrev $ratio
    fi
  
  #
  # for multiprocessing, need to make separate copy of the parameters file
  # so they don't conflict with each other
  
  cp $destdir/imexpr.par   ${tmp}imexpr.${seqno}.par
  cp $destdir/imstack.par  ${tmp}imstack.${seqno}.par

  PFILES=$tmp  \
     imexpr @@${tmp}imexpr.${seqno}.par \
        'a - b * c' $outfile $HDRFILE $Skyabbrev $ratio verbose=no mode=q \
        1>/dev/null
  exitval=$?
  
  PFILES=$tmp  \
    imstack @@${tmp}imstack.${seqno}.par \
       "${outfile},${HDRFILE},${Skyabbrev}" $tmpdestfile mode=q
# /data/oir/wcs/bin.solaris/imstack -o $tmpdestfile $outfile $HDRFILE $Skyabbrev
  exitval=$?

  /bin/rm $outfile; xit=$?

#  if [[ $extn -eq ".imh" ]] then
#    /bin/rm $(basename $outfile .imh).pix; xit=$(($xit | $?))
#    fi

  if (( xit != 0 )) then
    print "$(basename $0): can't rm ${outfile}\?\n"
    exit $xit
    fi
  
  delhead BANDID1 BANDID2 BANDID3 BANDID4 $tmpdestfile
  sethead BANDID1='spectrum - scaled sky'                        \
          BANDID2='raw spectrum'                                 \
          BANDID3='averaged sky'                                 \
          NSKYFIB=$Nskyfibers                                    \
          SKYFIB="$Skyfibers"                                    \
          SKYAPER="$Apertures"                                   \
          SKYCERT=$skycert                                       \
          NSKYMAX=$nskys                                         \
          SKYLINE=$emline                                        \
          SKYDIST=$fiberdist                                     \
          SKYMETHD="$skymethod"                                  \
          SKYRATIO=$ratio                                        \
          $tmpdestfile 
  
  /bin/mv $tmpdestfile $destfile; xit=$?

#  if (( $extn == ".imh" )) then
#    /bin/mv $(basename $tmpdestfile .imh).pix $(basename $destfile $extn).pix
#    xit=$(($xit | $?))
#    fi

  if (( xit != 0 )) then
    print "$(basename $0): can't mv ${outfile}\?\n"
    exit $xit
    fi
  
  if (( exitval != 0 )) then
    print -u2 -f "ERROR: imexpr: status $exitval\n" 
    fi
  if (( verbose > 1 )) then 
    printf "Output: %s\n\n" $outfile
    fi
  
  exit $exitval
 ) &  # END FORKED SUBSHELL

#mv $tmp /tmp/skyproc

  bkgnd[$nxtjob]=$!
  hdrfile[$nxtjob]=$HDRFILE
  nxtjob=$(($nxtjob+1)) 

# if mpjobs jobs are not running, continue forking before waiting
  if (( nxtjob - waitjob < mpjobs )) then
    continue
    fi
  
  wait ${bkgnd[$waitjob]}
  retval=$?

  retarr[$waitjob]=$retval
  if (( retval != 0 )) then
     printf "\nERROR val %d while skyproc.sub of %s\n" \
             $retval ${hdrfile[$waitjob]}
     exitval=$retval
     fi
  waitjob=$(($waitjob + 1))
  cat ${tmp}logfile.$waitjob
  /bin/rm -f ${tmp}logfile.$waitjob
done

# wait for last several to finish
while (( waitjob < nxtjob )) do
  wait ${bkgnd[$waitjob]}
  retval=$?
#  printf "Wait: %d %d ret = %d\n" $waitjob ${bkgnd[$waitjob]} $retval 
  retarr[$waitjob]=$retval
  if (( retval != 0 )) then
     printf "\nERROR val %d while skyproc.sub of %s\n"   \
          $retval ${hdrfile[$waitjob]}
     exitval=$retval
     fi
  waitjob=$(($waitjob + 1))
  cat ${tmp}logfile.$waitjob
done  

if (( verbose > 0 )) then
    if (( $logfile != "" )) then
	exec 1>> $logfile
	fi
    printf "\nElapsed time at exit: %s seconds\n" $SECONDS
    times
    fi

exit $exitval

# Feb 13 2004	DJM: Add option to use any of 4 sky emission lines with -e
# Feb 13 2004	DJM: Add option to use line width, height, or eq.width with -w
