# T_EREF -- Read reference features, shift, and stretch
# July 22, 2003
# By Doug Mink

# EREF reads a file produced by XCSAO or EMSAO containing the results of
# correlations between a template and a spectrum which is the same as the
# template but with distortions or the results of fits to emission lines
# in a spectrum.  It then fits the wavelength shifts to the wavelength from
# the start of the file and returns the power series coefficients.

include <pkg/gtools.h>
include <math/curfit.h>
include "identify.h"

define  MAXPTS          128

procedure t_reref

char	objfile[SZ_FNAME]	# Correlated file name
char	refin[SZ_FNAME]		# Input reference image
char	refout[SZ_FNAME]	# Output reference image
char	shiftfile[SZ_FNAME]	# XCSAO output file containing pixels and shifts
char	dbfile[SZ_FNAME]	# Path to id database file

bool	verbose, debug
int	i, nfeatures
int	ap[2]
double	dindef, dline0
char	s1[32], s2[32]
real	d1, d2, d3, d4
double	x[MAXPTS], y[MAXPTS], wts[MAXPTS], r[MAXPTS]
double	rmin			# Minimum R value for use of point in fit
int	ir

char	fitfunc[SZ_FNAME]	# Continuum fitting function type
char	logfile[SZ_FNAME]	# Log file
int	fitorder		# order of fit
int	niterate		# number of iterations for fit
real	grow			# radius around rejected points to replace
bool	interact		# Use interactive fitting (true or false)
int	npts
real	hirej, lowrej
real	smin
int	fd, logfd
pointer	cv
pointer	ic
pointer	gopen(), gt_init()
int     fscan(), open()

int	clgeti()
real	clgetr()
bool	clgetb()
double	clgetd()
double	dcveval()
pointer	id		# ID pointer
pointer	gt, gp
int	strsearch()

begin
	verbose = clgetb ("verbose")
	debug = clgetb ("debug")

	call clgstr ("object", objfile, SZ_FNAME)
	if (debug) {
	    call printf ("REREF: Crosscorrellated file is %s\n")
		call pargstr (objfile)
	    call flush (STDOUT)
	    }

	call strcpy ("database/rf", logfile, SZ_FNAME)
	call strcat (objfile, logfile, SZ_FNAME)
	if (debug) {
	    call printf ("REREF: Log file is %s\n")
		call pargstr (logfile)
	    call flush (STDOUT)
	    }
	iferr {logfd = open (logfile, NEW_FILE, TEXT_FILE)} then {
	    call delete (logfile)
	    logfd = open (logfile, NEW_FILE, TEXT_FILE)
	    }

        dindef = INDEFD
	call amovkd (0.d0, wts, MAXPTS)

	rmin = clgetd ("rmin")
	smin = clgetr ("smin")

#  Read X and shift vectors from XCSAO output
	call clgstr ("shiftfile", shiftfile, SZ_FNAME)
	fd = open (shiftfile, READ_ONLY, TEXT_FILE)
	if (debug) {
	    call printf ("REREF: Reading from shift file %s\n")
		call pargstr (shiftfile)
	    call flush (STDOUT)
	    verbose = YES
	    }
	npts = 0
	if (fd >0) {
	    while (fscan(fd) != EOF && npts < MAXPTS) {
		npts = npts + 1;
		call gargwrd (s1)
		call gargwrd (s2)
		call gargd (r[npts])
		call gargd (y[npts])
		call gargr (d1)
		call gargr (d2)
		call gargr (d3)
		call gargr (d4)
		call gargd (x[npts])
		if (r[npts] < rmin) {
		    wts[npts] = 0.0
		    }
		else {
		    wts[npts] = 1.0
		    }
		if (verbose) {
		    call printf ("%3d   %7.2f  %7.3f  %7.3f\n")
			call pargi (npts)
			call pargd (x[npts])
			call pargd (y[npts])
			call pargd (wts[npts])
		    }
		}
	    call close (fd)
	    if (verbose)
		call flush (STDOUT)
	    }

# Set the ICFIT pointer structure.
	call clgstr ("sfitfunc", fitfunc, SZ_FNAME)
	fitorder = clgeti ("sfitorder")
	niterate = 5
	hirej = smin
	lowrej = smin
	grow = 0
        call ic_open (ic)
        call ic_pstr (ic, "sample", "*")
        call ic_pstr (ic, "function", fitfunc)
        call ic_puti (ic, "naverage", 1)
        call ic_puti (ic, "order", fitorder)
        call ic_puti (ic, "niterate", niterate)
        call ic_putr (ic, "low", lowrej)
        call ic_putr (ic, "high", hirej)
        call ic_putr (ic, "grow", grow)
        call ic_pstr (ic, "ylabel", "")

#  Fit a function to X and Y (pixel and shift)
	if (verbose) {
	    call printf ("REREF: %d points to fit\n")
		call pargi (npts)
	    call flush (STDOUT)
	    }

#  If the interactive flag is set then use icg_fit to set the fitting
#  parameters.  Only done if task is run interactively as well.
	interact = clgetb ("sinteract")
	if (interact) {
	    iferr {
		gp = gopen ("stdgraph", NEW_FILE, STDGRAPH)
		gt = gt_init ()
		call gt_sets (gt, GTTYPE, "line")
		} then
		    call error (0, "Error opening `stdgraph'.")
	    call icg_fitd (ic,gp,"cursor",gt,cv,x,y,wts,npts)
	    call gt_free (gt)
	    call gclose (gp)
	    }

#  Do the fit non-interactively using ic_fit
	else {
	    call ic_fitd (ic,cv,x,y,wts,npts,YES,YES,YES,YES)
	    }

#  Convert coefficients to power series and print them
	function = dcvstati (ID_CV(id), CVTYPE)
	ncoeff = dcvstati (ID_CV(id), CVNCOEFF)
	call printf ("REREF: Function is %d, no. coeffs is %d\n")
	    call pargi (function)
	    call pargi (ncoeff)
	call dcvpower (ID_CV(id), pscoeff, ncoeff)
	call fprintf (fd, "Power Series Coefficients: ")
	do i = 1, ncoeff {
	    call fprintf (fd, " %.6f")
		call pargd (pscoeff[i])
	    }
	call fprintf (fd, "\n")


end

# Jul 21 2003	New program based on t_reref.x
